local utf8 = require("utf8")
local fps = require("fps")
local story = require("story")

local canvas

local shader

-- the font and love2d text object that will be used to display terminal content
local noto, text

local charset = "ﾘﾚﾌﾉﾐｺｲｳｨﾁﾜｦｿｻﾙﾆﾘﾚﾌﾉﾐｺｲｳｨﾁﾜｦｿｻﾙﾆﾘﾚﾌﾉﾐｺｲｳｨﾁﾜｦｿｻﾙﾆABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
local char_table
-- Populate the character table
char_table = {}
for p, c in utf8.codes(charset) do
  table.insert(char_table, utf8.char(c))
end

local trails = {}
-- this variable accumulates time that was spent on a frame, but not used
-- to spawn a trail. Say that a trail should spawn every 0.25 seconds. The last
-- frame took 0.4 seconds. That means that one trail should have spawned in the
-- duration of that frame. The remaining 0.4 - 0.25 = 0.15 seconds will be
-- stored in this variable, and might be used up in a successive frame to spawn
-- a new trail.
local trail_time_carry = 0

local pending = {}

local statics = {}
local target = {}

local rand

local logic_cr

-- table used to collect information about pressed keys
local keys = {any = {presses = 0, acc = {}}}

local state = {
  cursor_x = 1,
  cursor_y = 1,
  line_start = 1,
}

local config = {
  font_size = 21,
  speed = 8, -- downwards velocity, in chars/s
  chance_mutate = .005, -- chance to mutate an existing character
  col_spacing =  5, -- extra spacing between columns. can be negative
  row_spacing = -7, -- extra spacing between rows. can be negative
  min_trail_length = 10,
  max_trail_length = 35,
  background_color = {7, 14, 8},
  font_color = {77, 214, 79},
  font_color_highlight = {171, 212, 171},
  fade = 10, -- steps in alpha that fade per char in trail
  static_glitch_chance = 0.02,
  static_glitch_duration = 1,
  static_max_afterlife = 3,
  trails_per_s = 60, -- attempts to spawn a new trail, per second
  scale = 1,
}

-- rows and columns on the screen. will be updated when the window size changes
local cols = 0
local rows = 0
-- offset in x and y to center terminal content on screen. will be updated when
-- the window size changes
local off_x = 0
local off_y = 0

-- glyph dimensions. will be updated when the font is created
local glyph_width = -1
local glyph_height = -1

-- font used to print debug information
local debug_font

local function random_char()
  local char_index = rand:random(1, #char_table)
  return char_table[char_index]
end

local function print_char(x, y, char, color)
  if x > 0 and x <= cols and y > 0 and y <= rows then
    text:add({color, char},
             off_x + (x - 1) * (glyph_width + config.col_spacing),
             off_y + (y - 1) * (glyph_height + config.row_spacing))
  end
end

--[[
 ____ _____  _  _____ ___ ____ ____
/ ___|_   _|/ \|_   _|_ _/ ___/ ___|
\___ \ | | / _ \ | |  | | |   \___ \
 ___) || |/ ___ \| |  | | |___ ___) |
|____/ |_/_/   \_\_| |___\____|____/

--]]

local function new_static(posx, posy, char)
  local static = {
    posx = posx,
    posy = posy,
    char = char,
    current_char = char,
    glitch_rem = 0,
    dead = false,
  }

  function static.update(static, dt)
    -- if glitched, reduce remaining glitch time
    if static.glitch_rem > 0 then
      static.glitch_rem = static.glitch_rem - dt
      -- return to normal if glitch ended
      if static.glitch_rem <= 0 then
        static.current_char = static.char
      end
    -- else: check if glitch
    elseif rand:random() < config.static_glitch_chance * dt then
      static.current_char = random_char()
      static.glitch_rem = config.static_glitch_duration
    end

    -- check if this is still a target character
    if not target[static.posy] or not target[static.posy][static.posx] then
      if not static.afterlife then
        static.afterlife = rand:random() * config.static_max_afterlife
      end
    end

    if static.afterlife then
      static.afterlife = static.afterlife - dt
      if static.afterlife <= 0 then
        static.dead = true
      end
    end
  end

  function static.draw(static)
    print_char(static.posx, static.posy, static.current_char,
               config.font_color_highlight)
  end

  return static
end

local function create_static(px, py, char)
  if not statics[py] then
    statics[py] = {}
  end
  statics[py][px] = new_static(px, py, char)
end

--[[
 _____ ____      _    ___ _     ____
|_   _|  _ \    / \  |_ _| |   / ___|
  | | | |_) |  / _ \  | || |   \___ \
  | | |  _ <  / ___ \ | || |___ ___) |
  |_| |_| \_\/_/   \_\___|_____|____/

--]]

local function new_trail(length, posx)
  local trail = {
    length = length,
    chars = {},
    posx = posx,
    posy = -1,
    age = 0,
  }

  for i=1,trail.length do
    trail.chars[i] = random_char()
  end

  function trail.update(trail, dt)
    trail.age = trail.age + dt
    local new_posy = math.floor(trail.posy + trail.age * config.speed)
    local d = new_posy - trail.posy
    if d > 0 then
      trail.age = trail.age - d / config.speed
      trail.posy = new_posy
      for i=1,d do
        table.remove(trail.chars)
        table.insert(trail.chars, 1, random_char())
      end
    end

    -- check if the head of the tail occupies a target
    local px, py = trail.posx, trail.posy - 1
    if target[py] and target[py][px] then
      -- clear the char at this position in this trail so that it doesn't
      -- interfere with the static
      trail.chars[1] = ""
      local char = target[py][px]

      -- if there is no static yet for this target, create one
      if not statics[py] or not statics[py][px] or
         char ~= statics[py][px].char
      then
        create_static(px, py, char)
      end
    end

    for i=1,trail.length do
      if rand:random() < config.chance_mutate / i and trail.chars[i] ~= "" then
        trail.chars[i] = random_char()
      end
    end
  end

  function trail.draw(trail)
    for i,char in ipairs(trail.chars) do
      local r, g, b
      if i == 1 then
        r, g, b = unpack(config.font_color_highlight)
      else
        r, g, b = unpack(config.font_color)
      end
      local a = math.max(0, 255 - config.fade * i)
      print_char(trail.posx, trail.posy - i, char, {r, g, b, a})
    end
  end

  function trail.is_visible(trail)
    return trail.posx > 0 and trail.posx <= cols and
           (trail.posy - trail.length) <= rows
  end

  return trail
end

--[[
 ____ _____ ___  ______   __  ___ _   _ _____ _____ ____  _____ _    ____ _____
/ ___|_   _/ _ \|  _ \ \ / / |_ _| \ | |_   _| ____|  _ \|  ___/ \  / ___| ____|
\___ \ | || | | | |_) \ V /   | ||  \| | | | |  _| | |_) | |_ / _ \| |   |  _|
 ___) || || |_| |  _ < | |    | || |\  | | | | |___|  _ <|  _/ ___ \ |___| |___
|____/ |_| \___/|_| \_\|_|   |___|_| \_| |_| |_____|_| \_\_|/_/   \_\____|_____|

--]]

local function clear_text()
  target = {}
end

local function new_line()
  state.cursor_y = state.cursor_y + 1
  state.cursor_x = state.line_start
end

local function add_text(text)
  target[state.cursor_y] = target[state.cursor_y] or {}
  for char in string.gmatch(text, ".") do
    if char == "\n" then
      new_line()
      target[state.cursor_y] = target[state.cursor_y] or {}
    else
      target[state.cursor_y][state.cursor_x] = char
      state.cursor_x = state.cursor_x + 1
    end
  end
end

local function set_cursor(x, y)
  if x then
    state.cursor_x = x
    state.line_start = x
  end
  if y then
    state.cursor_y = y
  end
end

local function set_text(text, x, y)
  set_cursor(x, y)
  add_text(text)
end

local function clear_line(y, y_max)
  for i=y,(y_max or y) do
    target[i] = {}
  end
end

local function wait(seconds)
  local rem = seconds
  while rem > 0 do
    local dt = coroutine.yield()
    rem = rem - dt
  end
end

local function wait_for_key(valid_keys)
  local key = nil
  while not key do
    coroutine.yield()
    if #keys.any.acc > 0 then
      if valid_keys then
        for _, k in ipairs(keys.any.acc) do
          for _,valid in ipairs(valid_keys) do
            if valid == k then return k end
          end
        end
      else
        key = keys.any.acc[1]
      end
    end
  end
  return key
end

local function quit()
  clear_text()
  config.trails_per_s = 0
  local wait_time = 0
  if config.speed > 0 then
    wait_time = (rows + config.max_trail_length) / config.speed
  end
  wait_time = math.max(wait_time, config.static_max_afterlife)
  wait(wait_time + 1)
  love.event.push("quit")
end

local interface = {
  rows = function() return rows end,
  cols = function() return cols end,
  wait = wait,
  wait_for_key = wait_for_key,
  set_text = set_text,
  add_text = add_text,
  set_cursor = set_cursor,
  new_line = new_line,
  clear_text = clear_text,
  clear_line = clear_line,
  quit = quit,
  config = config,
}

--[[
 _     _____     _______   _____ _   _ _   _  ____ _____ ___ ___  _   _ ____
| |   / _ \ \   / / ____| |  ___| | | | \ | |/ ___|_   _|_ _/ _ \| \ | / ___|
| |  | | | \ \ / /|  _|   | |_  | | | |  \| | |     | |  | | | | |  \| \___ \
| |__| |_| |\ V / | |___  |  _| | |_| | |\  | |___  | |  | | |_| | |\  |___) |
|_____\___/  \_/  |_____| |_|    \___/|_| \_|\____| |_| |___\___/|_| \_|____/

--]]

local function set_size(w, h)
  w = w / config.scale
  h = h / config.scale
  canvas = love.graphics.newCanvas(w, h)

  cols = math.floor((w + config.col_spacing) / (glyph_width + config.col_spacing))
  off_x = math.floor((w + config.col_spacing - (cols * (glyph_width + config.col_spacing)))/2)

  rows = math.floor((h + config.row_spacing) / (glyph_height + config.row_spacing))
  off_y = math.floor((h + config.row_spacing - (rows * (glyph_height + config.row_spacing)))/2)

  shader:send("height", h)
end

function love.load()
  love.window.setTitle("[PROTOCOL #081]")

  debug_font = love.graphics.newFont(14)
  noto = love.graphics.newFont("NotoSansMonoCJKjp-Bold-Reduced.otf", config.font_size)
  text = love.graphics.newText(noto)
  canvas = love.graphics.newCanvas(1, 1)
  shader = love.graphics.newShader("shader.frag")

  glyph_height = noto:getHeight()
  glyph_width = noto:getWidth("m") -- for monospace fonts this is fine

  rand = love.math.newRandomGenerator()

  -- create coroutine for the story
  assert(type(story), "The story module must return a function")
  setfenv(story, setmetatable({},{
    __index = function(tab, key)
        return rawget(interface, key) or _G[key]
      end,
    }))

  logic_cr = coroutine.create(story)

  local w, h = love.window.getMode()
  set_size(w, h)

end

function love.update(dt)
  fps:update(dt)
  if logic_cr and coroutine.status(logic_cr) ~= "dead" then
    local succ, err = coroutine.resume(logic_cr, dt)
    if not succ then error(err) end
  end
  local num_remaining_trails = 0
  local remaining_trails = {}
  for _,trail in ipairs(trails) do
    trail:update(dt)
    if trail:is_visible() then
      num_remaining_trails = num_remaining_trails + 1
      remaining_trails[num_remaining_trails] = trail
    end
  end
  trails = remaining_trails

  for y,static_rows in pairs(statics) do
    for x, static in pairs(static_rows) do
      static:update(dt)
      if static.dead then
        statics[y][x] = nil
      end
    end
  end

  -- update the table that tracks pending trails per column
  for col,count in pairs(pending) do
    local rem = pending[col] - dt * config.speed
    -- this will clear this field as soon as there are no more pending characters
    pending[col] = rem > 0 and rem or nil
  end

  -- create new trails in random columns
  trail_time_carry = trail_time_carry + dt
  local trails_this_frame = math.floor(trail_time_carry * config.trails_per_s)
  local delta = trail_time_carry * config.trails_per_s - trails_this_frame
  trail_time_carry = delta / config.trails_per_s

  for i=1,trails_this_frame do
    local col = rand:random(1, cols)
    if not pending[col] then
      local len = rand:random(config.min_trail_length, config.max_trail_length)
      table.insert(trails, new_trail(len, col))
      pending[col] = len
    end
  end

  -- reset number of presses in keys table
  for key, tab in pairs(keys) do
    tab.presses = 0
  end
  keys.any.acc = {}
end

function love.resize(w, h)
  set_size(w, h)
end

function love.draw()
  text:clear()
  love.graphics.setCanvas(canvas)

  love.graphics.clear(config.background_color)
  love.graphics.setFont(noto)

  for _,trail in ipairs(trails) do
    trail:draw()
  end

  for _,static_rows in pairs(statics) do
    for _, static in pairs(static_rows) do
      static:draw()
    end
  end

  love.graphics.setColor(255, 255, 255)
  love.graphics.draw(text, 0, 0)

  love.graphics.setCanvas()
  love.graphics.setShader(shader)
  love.graphics.draw(canvas, 0, 0, 0, config.scale, config.scale)
  love.graphics.setShader()

  love.graphics.setFont(debug_font)
  love.graphics.setColor(unpack(config.font_color_highlight))
  local debug_string = string.format("%sfps\t%dx%d\t%d trails",
                                     fps.fps,
                                     rows, cols,
                                     #trails)
  -- love.graphics.print(debug_string, 0, 0)

end

function love.keypressed(key)
  keys[key] = keys[key] or {presses = 0, pressed = fase}
  keys[key].presses = keys[key].presses + 1
  keys[key].pressed = true
  keys.any.presses = keys.any.presses + 1
  table.insert(keys.any.acc, key)
end

function love.keyreleased(key)
  keys[key] = keys[key] or {presses = 0, pressed = fase}
  keys[key].pressed = false
end
